/*
 * Implementation of Federal Information Processing Standards Publication
 * FIPS 180-1 (17 Apr 1995) which supersedes FIPS 180 (11 May 1993)
 *
 * Speed hack version optimised for speed (see also reference version)
 *
 * Adam Back <aba@dcs.ex.ac.uk>
 *
 */

#include "sha1.h"
#include "endian.h"

#define min( x, y ) ( ( x ) < ( y ) ? ( x ) : ( y ) )

/********************* function used for rounds 0..19 ***********/

/* #define F1( B, C, D ) ( ( (B) & (C) ) | ( ~(B) & (D) ) ) */

/* equivalent, one less operation: */
#define F1( B, C, D ) ( (D) ^ ( (B) & ( (C) ^ (D) ) ) )


/********************* function used for rounds 20..39 ***********/

#define F2( B, C, D ) ( (B) ^ (C) ^ (D) )

/********************* function used for rounds 40..59 ***********/

/* #define F3( B, C, D ) ( (B) & (C) ) | ( (C) & (D) ) | ( (C) & (D) ) */

/* equivalent, one less operation */

#define F3( B, C, D ) ( ( (B) & ( (C) | (D) )) | ( (C) & (D) ) )

/********************* function used for rounds 60..79 ***********/

#define F4( B, C, D ) ( (B) ^ (C) ^ (D) )

#define K1 0x5A827999  /* constant used for rounds 0..19 */
#define K2 0x6ED9EBA1  /* constant used for rounds 20..39 */
#define K3 0x8F1BBCDC  /* constant used for rounds 40..59 */
#define K4 0xCA62C1D6  /* constant used for rounds 60..79 */

/* magic constants */

#define H0 0x67452301
#define H1 0xEFCDAB89
#define H2 0x98BADCFE
#define H3 0x10325476
#define H4 0xC3D2E1F0

word32 H[ 5 ] = { H0, H1, H2, H3, H4 };

/* rotate X n bits left   ( X <<< n ) */

#define S(n, X) ( ( (X) << (n) ) | ( (X) >> ( 32 - (n) ) ) )

/* convert to big endian */

static void convert_to_big( void* data, int len )
{
   word32 temp;
   byte* temp_as_bytes = (byte*) &temp;
   word32* data_as_words = (word32*) data;
   byte* data_as_bytes;
   int i;

   len /= 4;
   for ( i = 0; i < len; i++ )
   {
      temp = data_as_words[ i ];
      data_as_bytes = (byte*) &( data_as_words[ i ] );
      
      data_as_bytes[ 0 ] = temp_as_bytes[ 3 ];
      data_as_bytes[ 1 ] = temp_as_bytes[ 2 ];
      data_as_bytes[ 2 ] = temp_as_bytes[ 1 ];
      data_as_bytes[ 3 ] = temp_as_bytes[ 0 ];
   }
}

/* do this with a macro as little_endian may be compile time, to avoid
   function call overhead */

#define convert_to_bigendian( data, len ) \
   ( little_endian ? convert_to_big( data, len ) : 0 )

void SHA1_init( SHA1_ctx* ctx )
{
   memcpy( ctx->H, H, 5 * sizeof( word32 ) );
#if defined( word64 )
   ctx->bits = 0;
#else
   ctx->lbits = 0;
   ctx->hbits = 0;
#endif
   ctx->mlen = 0;
}

static void SHA1_transform( SHA1_ctx* ctx )
{
   int t;
   word32 A = ctx->H[ 0 ];
   word32 B = ctx->H[ 1 ];
   word32 C = ctx->H[ 2 ];
   word32 D = ctx->H[ 3 ];
   word32 E = ctx->H[ 4 ];
   word32* W = (word32*) ( ctx->M );

/* Use method B from FIPS-180 (see fip-180.txt) where the use of
   temporary array W of 80 word32s is avoided by working in a circular
   buffer of size 16 word32s.  

   Reuse the 16 word32 array formed by the 64 bytes of the message in
   the context struct: W = ctx->M */

/********************* define some macros *********************/

/* Wc = access W as 16 word circular buffer */

#define Wc( t ) ( W[ (t) & 15 ] )

/* Calculate access to W array on the fly for entries 16 .. 79 */

#define Wf( t ) \
   ( Wc( t ) = S( 1, Wc( t ) ^ Wc( t - 14 ) ^ Wc( t - 8 ) ^ Wc( t - 3 ) ) )

/* Calculate access to W virtual array calculating access to W on the fly */

#define Wfly( t ) ( (t) < 16 ? W[ (t) ] : Wf( (t) ) )

#define ROUND( t, A, B, C, D, E, Func, K ) \
   E += S( 5, A ) + Func( B, C, D ) + Wfly( t ) + K;\
   B = S( 30, B );

/* Remove rotatation E' = D; D' = C; C' = B; B' = A; A' = E; by
   completely unrolling and rotating the arguments to the macro ROUND
   manually so the rotation is compiled in.
*/

#define ROUND5( t, Func, K ) \
   ROUND( t + 0, A, B, C, D, E, Func, K );\
   ROUND( t + 1, E, A, B, C, D, Func, K );\
   ROUND( t + 2, D, E, A, B, C, Func, K );\
   ROUND( t + 3, C, D, E, A, B, Func, K );\
   ROUND( t + 4, B, C, D, E, A, Func, K )

#define ROUND20( t, Func, K )\
   ROUND5( t +  0, Func, K );\
   ROUND5( t +  5, Func, K );\
   ROUND5( t + 10, Func, K );\
   ROUND5( t + 15, Func, K )

/********************* use the macros *********************/

/* rounds  0..19 */

   ROUND20(  0, F1, K1 );

/* rounds 21..39 */

   ROUND20( 20, F2, K2 );

/* rounds 40..59 */

   ROUND20( 40, F3, K3 );

/* rounds 60..79 */

   ROUND20( 60, F4, K4 );

   ctx->H[ 0 ] += A;
   ctx->H[ 1 ] += B;
   ctx->H[ 2 ] += C;
   ctx->H[ 3 ] += D;
   ctx->H[ 4 ] += E;
}

void SHA1_update( SHA1_ctx* ctx, const void* vdata, word32 data_len )
{
   const byte* data = vdata;
   word32 use;
   word32 low_bits;
   
/* convert data_len to bits and add to the 64 bit word formed by lbits
   and hbits 
*/

#if defined( word64 )
   ctx->bits += ( (word64) data_len ) << 3;
#else
   ctx->hbits += data_len >> 29;
   low_bits = data_len << 3;
   ctx->lbits += low_bits;
   if ( ctx->lbits < low_bits ) { ctx->hbits++; }
#endif

/* deal with first block */

   use = min( 64 - ctx->mlen, data_len );
   memcpy( ctx->M + ctx->mlen, data, use );
   ctx->mlen += use;
   data_len -= use;
   data += use;

   while ( ctx->mlen == 64 )
   {
      convert_to_bigendian( (word32*)ctx->M, 64 );
      SHA1_transform( ctx );
      use = min( 64, data_len );
      memcpy( ctx->M, data, use );
      ctx->mlen = use;
      data_len -= use;
   }
}

void SHA1_final( SHA1_ctx* ctx )
{
#if defined( word64 )
   word64 temp;
#endif

   if ( ctx->mlen < 56 )
   {
      ctx->M[ ctx->mlen ] = 0x80; ctx->mlen++;
      memset( ctx->M + ctx->mlen, 0x00, 56 - ctx->mlen );
      convert_to_bigendian( ctx->M, 56 );
   }
   else
   {
      ctx->M[ ctx->mlen ] = 0x80; 
      ctx->mlen++;
      memset( ctx->M + ctx->mlen, 0x00, 64 - ctx->mlen );
      convert_to_bigendian( ctx->M, 64 );
      SHA1_transform( ctx );
      memset( ctx->M, 0x00, 56 );
   }

#if defined( word64 )
   temp = ( ctx->bits << 32 | ctx->bits >> 32 );
   memcpy( ctx->M + 56, &temp, 8 );
#else
   memcpy( ctx->M + 56, &(ctx->hbits), 8 );
#endif
   SHA1_transform( ctx );
}

void SHA1_digest( SHA1_ctx* ctx, byte* digest )
{
   if ( digest )
   {
      memcpy( digest, ctx->H, 5 * sizeof( word32 ) );
      convert_to_bigendian( digest, 5 * sizeof( word32 ) );
   }
}
